"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const bluebird = require("bluebird");
const crypto = require("crypto");
const dns = require("dns");
const EventEmitter = require("events");
const fs = require("fs");
const jaeger_client_1 = require("jaeger-client");
const kafka_node_1 = require("kafka-node");
const Nats = require("nats");
const redis = require("redis");
const util_1 = require("util");
const worker_threads_1 = require("worker_threads");
const logs_1 = require("./logs");
const logs_threads_1 = require("./logs_threads");
const threads_1 = require("./threads");
const types_1 = require("./types");
const util_2 = require("./util");
const production_1 = require("./production");
bluebird.promisifyAll(redis.RedisClient.prototype);
bluebird.promisifyAll(redis.Multi.prototype);
/**
 *  Clase para las conexiones y publicaciones sobre
 *  la cola de mensajes.
 */
class Brokers extends util_2.default {
    /**
     * Instancia de la implementacion del http transport
     *
     * @private
     * @type {HttpTransport}
     * @memberof Brokers
     */
    // private httpTransportInstance: HttpTransport;
    /**
     * Contructor de la clase.
     * @param kafkaHost Servidores Kafka
     * @param Topic Topico de escucha
     */
    constructor(Conf) {
        //  Contructor padre
        super();
        /**
         * ID unico de la instancia, cada que reinicia el servicio,
         * el ID cambia.
         *
         * @protected
         * @type {string}
         * @memberof Brokers
         */
        this.uuid = null;
        /**
         * Manejador de eventos
         * @protected
         */
        this.event = new EventEmitter();
        /**
         * Manejador de logs
         * @protected
         */
        this.logs = logs_1.default;
        /**
         * Servidores de kafka
         * @protected
         */
        this.kafkaHost = [];
        /**
         * Hosts de los brokers
         * @protected
         */
        this.Hosts = [];
        /**
         * Variable para revisar si NATS se encuentra conectado o desconectado
         * @protected
         */
        this.natsIsConnected = false;
        /**
         * variable que almacena la direccion
         * actual del archivo que se ejecuta.
         * @private
         */
        this.cwd = '';
        /**
         * Indicador si el hilo en reserva se encuentra activo.
         * @private
         */
        this.threadResInit = false;
        /**
         * Instancia de la conexion a NATS
         * @private
         */
        this.connNats = null;
        /**
         * PID de la conexion a la subscripcion en
         * NATS
         * @private
         */
        this.pidSub = null;
        /**
         * Numero de hilos registrados actualmente
         * @private
         */
        this.threadTotal = 1;
        /**
         * Maximo numero de hilos que se soportan
         * @private
         */
        this.maxThread = 0;
        /**
         * Indicador para conocer si el consumer se encuentra en pause.
         * @private
         */
        this.isPauseConsumer = false;
        /**
         * Indicador del soporte de hilos
         * @private
         */
        this.threadSupport = true;
        // Validar si se envian los logs a Fluentd
        if (process.env.LOGS_FLUENT) {
            this.logs.config(Conf);
        }
        // Generar el id unico del servicio
        this.uuid = util_2.default.key();
        // Asignar la direccion actual
        this.cwd = process.cwd();
        // Verificar el soporte de hilos
        this.threadSupport = false; // !isUndefined(Conf.Thread) ? Conf.Thread : false;
        // Asignar el numero maximo de hilos disponibles
        this.maxThread = process.env.MAX_THREAD
            ? parseInt(process.env.MAX_THREAD, 0)
            : 100;
        // Crear el administrador de los Thread
        this.thread = new threads_1.default(process.cwd(), __dirname);
        // Asignar la configuracion
        this.conf = Conf;
        this.conf.ClearQueque = util_1.isUndefined(Conf.ClearQueque)
            ? true
            : Conf.ClearQueque;
        // Verificar si no se propociono un tipo de particion
        if (util_1.isUndefined(this.conf.PartitionerType) ||
            util_1.isNull(this.conf.PartitionerType)) {
            this.conf.PartitionerType =
                this.conf.PartitionerType || types_1.PartitionerType.RANDOM; // Default
        }
        /**
         * Definir topico con soporte para servicios con nomemclatura
         * legada.
         */
        if (this.conf.Version === 'v1' && this.conf.App === 'appcoppel') {
            // Es un servicio legado de appcoppel
            this.Topic = this.conf.Name;
        }
        else {
            // Crear la combinacion
            this.Topic = `${this.conf.App}_${this.conf.Version}_${this.conf.Name}`;
        }
        // Ver si corresponde a una conexion sobre NATS
        if (worker_threads_1.isMainThread) {
            if (process.env.REST_ENABLED) {
                this.logs.info('HTTP transport: enabled');
                // this.connecHttp();
            }
            else {
                // Validar si es NATS
                if (process.env.NATS) {
                    this.connecNats();
                }
                else {
                    // Kafka por defualt
                    this.connectKafka();
                }
            }
            if (this.threadSupport) {
                // Crear hilo
                this.threadRes = this.thread.execute({});
                this.threadRes.on('online', () => {
                    this.threadResInit = true;
                });
            }
            // Se verifica si hay configuracion de Redis para validacion de la informacion
            const conRedisValidate = process.env.REDISVALIDATE || 'redis-validate.logging';
            if (conRedisValidate === 'redis-validate.logging') {
                dns.lookup(conRedisValidate, (err, address, family) => {
                    if (!err) {
                        this.initializeRedis(`redis://${conRedisValidate}`);
                    }
                });
            }
            else {
                this.initializeRedis(conRedisValidate);
            }
        }
        else {
            // Es un hilo secundario
            worker_threads_1.parentPort.on('message', (data) => {
                // Publicar para su procesamiento
                this.message(data.Data);
            });
            // Cambiar el objeto de logs
            this.logs = new logs_threads_1.default();
        }
        this.initializeTracer();
    }
    /**
     * Metodo para validar conexion al broker
     */
    validateConnection() {
        // En caso de REST siempre regresa un true
        if (process.env.REST_ENABLED) {
            return true;
        }
        if (worker_threads_1.isMainThread) {
            const validateRedis = this.redisClient
                ? this.isRedisConnected
                : true;
            if (process.env.NATS) {
                return this.natsIsConnected && validateRedis;
            }
            else {
                if (!this.client) {
                    return false;
                }
                return new Promise((rr, ff) => {
                    this.client.topicExists([this.Topic], (error) => {
                        rr(error && validateRedis ? false : true);
                    });
                });
            }
        }
        return false;
    }
    /**
     * Metodo para realizar el envio de datos a la cola de mensajes
     * @param data Datos a enviar
     * @param topic Topico de publicacion
     * @param key Clave unica
     */
    send(dataMessage, topic, key = util_2.default.key()) {
        if (worker_threads_1.isMainThread) {
            if (process.env.REST_ENABLED) {
                // Convertir el string
                const resp = JSON.parse(Buffer.from(dataMessage, 'base64').toString('utf8'));
                // Emitir la respuesta del evento
                // this.httpTransportInstance.emit('response', resp);
            }
            else {
                if (process.env.NATS) {
                    // Responder la peticion
                    this.nc.publish(topic, dataMessage);
                }
                else {
                    this.producer.send([
                        {
                            key,
                            messages: [dataMessage],
                            topic,
                        },
                    ], (err) => {
                        if (err) {
                            // Imprimir el error
                            this.logs.error(err);
                        }
                    });
                }
            }
        }
        else {
            // Publicar en el hilo principal
            worker_threads_1.parentPort.postMessage({
                Data: {
                    dataMessage,
                    key,
                    topic,
                },
                Type: types_1.TypeMessage.Publish,
            });
        }
    }
    /**
     * Metodo para inicializar redis para validacion de mensajes
     */
    initializeRedis(redisCon) {
        this.redisClient = redis.createClient({
            enable_offline_queue: false,
            url: redisCon,
        });
        this.redisClient.on('ready', () => {
            this.logs.info(`Conexión realizada a redis [${redisCon}] revision de duplicados`);
            this.isRedisConnected = true;
        });
        this.redisClient.on('error', () => {
            this.logs.info(`Error en la conexión a redis de revision de duplicados`);
            this.isRedisConnected = false;
        });
    }
    /**
     * Metodo de inicializacion del Tracer
     */
    initializeTracer() {
        const JAEGER_AGENT_HOST = process.env.JAEGER_AGENT_HOST || 'jaeger-agent.logging';
        const version = JSON.parse(fs.readFileSync(`${__dirname}/../package.json`).toString()).version;
        const config = {
            reporter: {
                agentHost: JAEGER_AGENT_HOST,
            },
            sampler: {
                param: 1,
                type: 'const',
            },
            serviceName: this.Topic,
        };
        const options = {
            tags: {
                '@coppel-ts-version': `ts_${version}`,
            },
        };
        this.tracer = jaeger_client_1.initTracer(config, options);
    }
    /**
     * Metodo para iniciar el servidor del transport de HTTP
     *
     * @private
     * @memberof Brokers
     */
    connecHttp() {
        /* // Iniciar el servidor http
        this.httpTransportInstance = new HttpTransport(
            `${this.conf.App}_${this.conf.Version}_this.conf.Name`,
            this.conf.Version,
            this.conf.App,
            this.conf.Name,
        );

        // Conectar con los eventos de mensajes
        this.httpTransportInstance.on('message', (data: any) => {
            this.processMessage(data);
        });

        // Iniciar el servidor
        try {
            this.httpTransportInstance.start();
        } catch (error) {
            this.logs.error('Ocurrio un error al iniciar el http transport');
            this.logs.console.error(error);
            process.exit(-33);
        } */
    }
    /**
     * Metodo para realizar la conexion Nats
     * @private
     */
    connecNats() {
        // Verificar si se paso la cadena de conexion a kafka
        if (this.conf.Nats) {
            // Asignar la pasada como configuracion
            this.Hosts = this.conf.Nats;
        }
        else {
            // Verificar que se asigno una por variable de ambiente
            if (process.env.NATS) {
                // Recorrer todos los elementos para darle formato
                for (const i of process.env.NATS.split(',')) {
                    this.Hosts.push(`nats://${i}`); // Dar el formato correcto
                }
            }
            else {
                this.logs.error('No se proporciono la configuracion de NATS');
                process.exit(-1); // Finalizar el programa
            }
        }
        // Realizar la conexion a NATS - nc.currentServer.url.host
        this.nc = Nats.connect({ servers: this.Hosts, name: this.Topic });
        // Evento que indica la conexion
        this.nc.on('connect', (nc) => {
            const text = `Escuchando el topico ${this.Topic} en Nats ${this.nc.currentServer.url.host}`;
            // Notificar que la conexion fue exitosa.
            if (production_1.default) {
                this.logs.systemInfo(text);
            }
            else {
                this.logs.info(text);
            }
            // Almacenar la conexion
            this.connNats = nc;
            // Realizar la conexion de subcripcion al topico
            this.pidSub = this.connNats.subscribe(this.Topic, { queue: this.Topic }, this.handlerNats.bind(this));
            this.natsIsConnected = true;
        });
        this.nc.on('disconnect', () => {
            this.natsIsConnected = false;
        });
        this.nc.on('reconnect', () => {
            this.natsIsConnected = true;
        });
    }
    /**
     * Metodo para realizar la conexion al broker de kafka
     * @private
     */
    connectKafka() {
        let intentReconection = 0;
        const RQ = process.env.REQUEST_TIMEOUT;
        const requesTimeout = process.env.REQUEST_TIMEOUT
            ? RQ
            : 20000;
        // Verificar si se paso la cadena de conexion a kafka
        if (this.conf.Broker) {
            // Asignar la pasada como configuracion
            this.kafkaHost = this.conf.Broker;
        }
        else {
            // Verificar que se asigno una por variable de ambiente
            if (process.env.KAFKA) {
                // Asignar
                this.kafkaHost = process.env.KAFKA.split(',');
            }
            else {
                this.logs.error('No se proporciono la configuracion de KAFKA');
                process.exit(-1); // Finalizar el programa
            }
        }
        // Crear el cliente de kafka.
        this.client = new kafka_node_1.KafkaClient({
            kafkaHost: this.kafkaHost.join(','),
            requestTimeout: requesTimeout,
        });
        // Registrar el producer.
        this.producer = new kafka_node_1.Producer(this.client, {
            partitionerType: this.conf.PartitionerType,
        });
        // Eventos del producer
        this.producer.on('ready', () => {
            let consumer = this.Topic;
            if (process.env.CONSUMER_GROUP) {
                consumer = process.env.CONSUMER_GROUP;
            }
            else if (!util_1.isUndefined(this.conf.ConsumerGroup)) {
                consumer = this.conf.ConsumerGroup;
            }
            if (util_1.isUndefined(this.conf.Offset) || util_1.isNull(this.conf.Offset)) {
                this.conf.Offset = types_1.TypeOffset.Latest;
            }
            if (util_1.isUndefined(this.conf.OutOfRangeOffset) ||
                util_1.isNull(this.conf.OutOfRangeOffset)) {
                this.conf.OutOfRangeOffset = types_1.TypeOffset.Earliest;
            }
            // Opciones del consumidor
            const options = {
                fromOffset: this.conf.Offset,
                groupId: consumer,
                // tslint:disable-next-line: object-literal-sort-keys
                OutOfRangeOffset: this.conf.OutOfRangeOffset,
                kafkaHost: this.kafkaHost.join(','),
            };
            // Crear el consumidor
            this.consumer = new kafka_node_1.ConsumerGroup(options, [this.Topic]);
            // Escuchar cuando se realice la conexion
            this.consumer.on('connect', () => {
                const text = `Escuchando el topico ${this.Topic} en el Kafka ${this.kafkaHost.join(',')} - Consumer: ${consumer}`;
                // Notificar que la conexion fue exitosa.
                if (production_1.default) {
                    this.logs.systemInfo(text);
                }
                else {
                    this.logs.info(text);
                }
            });
            // Llegada de mensajes
            this.consumer.on('message', (message) => {
                try {
                    // Convertir el mensaje
                    const data = JSON.parse(Buffer.from(message.value.toString(), 'base64').toString('utf8'));
                    // Pasar el mensaje al que lo procesara
                    this.processMessage(data);
                }
                catch (error) {
                    // Disparamos la excepcion
                    this.exception(this.Topic, message.value.toString(), error);
                }
            });
            // Errores
            this.consumer.on('error', (err) => {
                if ((err && err.name === 'BrokerNotAvailableError') ||
                    err.name === 'NestedError') {
                    // Cantidad de reconexiones
                    intentReconection = intentReconection + 1;
                    this.logs.warn('attempting reconnect');
                    this.consumer.client.refreshMetadata(this.Topic.split(','), (error) => {
                        if (intentReconection >= 5) {
                            this.logs.error('[CONSUMER] - Se pasaron del maximo numero de reintentos de reconexion.');
                            this.logs.error(error);
                            process.exit(143);
                        }
                        else {
                            this.logs.warn(`[CONSUMER] - Ocurrio el siguiente error ${err.name}: ${err.message} durante la reconexion.`);
                        }
                    });
                }
                else {
                    this.logs.error(`[CONSUMER] - ${err.name}: ${err.message} - Se detentra el servicio.`);
                    // Enviar el error
                    this.logs.error(err.stack);
                    process.exit(143);
                }
            });
        });
        this.producer.on('error', (err) => {
            this.logs.error(`[PRODUCER] - ${err.name}: ${err.message} - Se detentra el servicio.`);
            // Enviar el error
            this.logs.error(err);
            process.exit(143);
        });
    }
    /**
     * Handler para la escucha de mensajes de NATS
     * @param msg Mensaje de entrada
     * @param reply Id del que solicita el dato
     * @param subject Topico
     */
    handlerNats(msg, reply, subject) {
        try {
            // Convertir el mensaje
            let data = JSON.parse(Buffer.from(msg, 'base64').toString('utf8'));
            // agregar el ID para el reply
            if (!data.uuid && !data.metadata.asynchronous) {
                data = Object.assign(data, { uuid: reply });
            }
            // Pasar el mensaje al que lo procesara
            this.processMessage(data);
        }
        catch (error) {
            // Disparamos la excepcion
            this.exception(this.Topic, msg, error);
        }
    }
    /**
     * Metodo para validar si se ha procesado previamente
     * la peticion.
     *
     * @private
     * @param {*} data
     * @returns {Promise<boolean>}
     * @memberof Brokers
     */
    hasProcessMessage(data) {
        // Se calcula el sha256 de la data
        // tslint:disable-next-line: max-line-length
        const sha256 = crypto.createHash('sha256').update(data).digest('hex');
        return new Promise((r, f) => __awaiter(this, void 0, void 0, function* () {
            // Se busca en redis si se encuentra la informacion
            this.redisClient.send_command('GETSET', [sha256, 'on'], (error, respAction) => {
                if (error) {
                    f(false);
                }
                else {
                    if (util_1.isNull(respAction)) {
                        r(true);
                    }
                    else {
                        r(false);
                    }
                }
            });
        }));
    }
    /**
     * Metodo para crear una instancia del servicio para
     * el procesamiento del mensaje.
     * @param data Datos que se envian a la instancia
     */
    processMessage(data) {
        return __awaiter(this, void 0, void 0, function* () {
            // Se valida si esta activado la validacion por redis
            if (this.redisClient) {
                try {
                    // tslint:disable-next-line: max-line-length
                    // Si se encuentra una respuesta, cancelamos que continue la peticion,
                    // ya que significa que ya se proceso
                    if (yield this.hasProcessMessage(`${this.Topic}_${JSON.stringify(data)}`) /*resp[0]*/) {
                        return;
                    }
                }
                catch (error) {
                    this.logs.error(error);
                    this.response(data, -97);
                    return;
                }
            }
            // Validar el soporte de hilos
            if (this.threadSupport) {
                const w = this.threadRes;
                this.threadRes = this.thread.execute(data);
                // Agregar un elemento
                this.threadTotal++;
                // Verificar si se pausa el consumer
                if (!this.isPauseConsumer &&
                    this.threadTotal >= this.maxThread - 1) {
                    // Iniciar de nuevo el consumer
                    this.pause();
                }
                // Evento de conexion
                this.threadRes.on('online', () => {
                    this.threadResInit = true;
                });
                // Validar si el hilo ya se encuentra activo
                if (this.threadResInit) {
                    w.postMessage({ type: 'message', Data: data });
                }
                else {
                    w.on('online', () => {
                        w.postMessage({ type: 'message', Data: data });
                    });
                }
                // Esperar los mensajes
                w.on('message', (response) => {
                    if (response.Type === types_1.TypeMessage.Publish) {
                        const dataResponse = response.Data;
                        // Publicar
                        this.send(dataResponse.dataMessage, dataResponse.topic, dataResponse.key);
                        // Eliminar el Hilo
                        w.terminate();
                        // Restar un elemento
                        this.threadTotal--;
                        // Ver si se requiere reanudar el consumer
                        if (this.isPauseConsumer &&
                            this.threadTotal < this.maxThread - 1) {
                            // Iniciar de nuevo el consumer
                            this.resume();
                        }
                    }
                    else if (response.Type === types_1.TypeMessage.Logs) {
                        const logs = response.Data;
                        if (logs.Type === types_1.TypeLogs.Info) {
                            this.logs.info(logs.Message);
                        }
                        else if (logs.Type === types_1.TypeLogs.Error) {
                            this.logs.error(logs.Message);
                        }
                        else if (logs.Type === types_1.TypeLogs.Warning) {
                            this.logs.warning(logs.Message);
                        }
                        else if (logs.Type === types_1.TypeLogs.Log) {
                            this.logs.log(logs.Message);
                        }
                    }
                });
                w.on('error', (error) => {
                    this.logs.error(error);
                });
                w.on('exit', (exitCode) => {
                    this.logs.info(`Worker ${w.threadId} has stopped with code ${exitCode}`);
                });
            }
            else {
                this.message(data);
            }
        });
    }
    /**
     * Metodo para pausar la llegada de mensajes
     * desde el broker
     * @private
     */
    pause() {
        if (this.conf.Nats) {
            this.connNats.drainSubscription(this.pidSub);
        }
        else {
            this.consumer.pause();
        }
        // Indicar que se esta en pause
        this.isPauseConsumer = true;
    }
    /**
     * Metodo para activar la escucha de mensajes
     * desde el broker.
     * @private
     */
    resume() {
        if (this.conf.Nats) {
            this.pidSub = this.connNats.subscribe(this.Topic, { queue: this.Topic }, this.handlerNats.bind(this));
        }
        else {
            this.consumer.resume();
        }
        // Indicar que sale de la pausa
        this.isPauseConsumer = false;
    }
}
exports.default = Brokers;
